/*****************************************************************************/
//
//  superposition_impl_nv.hpp: C++ version of Perl module "Quantum::Superpositions"
//      Implementation for non-variadic version using Boost PP iteration.
//
//   $URL: https://yak.myhome.cx/repos/source/mylib/trunk/cpp/superposition_impl_nv.hpp $
//  $Date: 2007-10-25 20:57:55 +0900 (Thu, 25 Oct 2007) $
//   $Rev: 480 $
//
//      Copyright (c) 2007 Yak!
//
//  This code is distributed under the zlib/libpng License.
//  See http://www.opensource.org/licenses/zlib-license.php
//
/*****************************************************************************/

#if !BOOST_PP_IS_ITERATING

#ifndef YAK_UTIL_SUPERPOSITION_IMPL_NV_HPP
#define YAK_UTIL_SUPERPOSITION_IMPL_NV_HPP

#include <boost/mpl/always.hpp>

#include <boost/tuple/tuple.hpp>

#include <boost/preprocessor/repetition/enum.hpp>
#include <boost/preprocessor/repetition/enum_binary_params.hpp>
#include <boost/preprocessor/repetition/enum_params.hpp>
#include <boost/preprocessor/iteration/iterate.hpp>

namespace yak {
namespace util {
namespace superposition {

#define BOOST_PP_ITERATION_LIMITS (1, 10)
#define BOOST_PP_FILENAME_1 "superposition_impl_nv.hpp"

#include BOOST_PP_ITERATE()

#undef BOOST_PP_ITERATION_LIMITS
#undef BOOST_PP_FILENAME_1

} // namespace superposition
} // namespace util
} // namespace yak

#endif // YAK_UTIL_SUPERPOSITION_IMPL_NV_HPP

#else // !BOOST_PP_IS_ITERATING

#define N BOOST_PP_ITERATION()

#define TYPE_TN BOOST_PP_ENUM_PARAMS(N, T)
#define VAR_TN BOOST_PP_ENUM_PARAMS(N, t)
#define TYPE_UN BOOST_PP_ENUM_PARAMS(N, U)
#define VAR_UN BOOST_PP_ENUM_PARAMS(N, u)
#define TYPENAME_TN BOOST_PP_ENUM_PARAMS(N, typename T)
#define TYPENAME_UN BOOST_PP_ENUM_PARAMS(N, typename U)
#define CONST_TN_REF_TN BOOST_PP_ENUM_BINARY_PARAMS(N, const T, &t)

#define holderN BOOST_PP_CAT(holder, N)

template<TYPENAME_TN>
class holderN;

namespace detail {

// Functor class for operator

// Though gcc/conceptgcc can handle even if just a bool is used, dependent name seems 
// to be necessary as required by, at least, VC 7.1/8. The reason is as follows:
// If we used just a bool, the RelOpN::result_type becomes non-dependent name.
// Inside OPERATOR_CLASS_N(n)::operator() definition, the result_type is used in the 
// context that requires the completeness of the type (construction).
// So, the completeness of the result_type affects whether ill-formed or well-formed of the
// definition of the operator(). At the point of definition, the result_type is incomplete.
// At the point of instantiation, the result_type is complete. This seems to violate 
// 14.6/7 in 14882:2003.
// By making the result_type dependent-name, we can avoid the violation.
#define BOOL(z, n, data) typename boost::mpl::always<bool>::template apply<T##n>::type

#define RelOpN BOOST_PP_CAT(RelOp, N)

template<typename U, TYPENAME_TN>
struct RelOpN
{
	typedef holderN<BOOST_PP_ENUM(N, BOOL, _)> result_type;
	RelOpN(spos_type type, const U& u) : type(type), arg(u) {}
	spos_type type;
	const U& arg;
};

#define OPERATOR_APPLY(z, i, n) (t##i OPERATOR_OP(n) this->arg)
#define OPERATOR_CLASS_N(n) BOOST_PP_CAT(OPERATOR_CLASS(n), N)

// template for functor class defniition
#define OPERATOR_TEMPLATE(z, n, data) \
template<typename U, TYPENAME_TN> \
struct OPERATOR_CLASS_N(n) : public RelOpN<U, TYPE_TN> \
{ \
	typedef typename RelOpN<U, TYPE_TN>::result_type result_type; \
	OPERATOR_CLASS_N(n)(spos_type type, const U& u) : RelOpN<U, TYPE_TN>(type, u) {} \
	result_type operator()(CONST_TN_REF_TN) \
	{ \
		return result_type(this->type, BOOST_PP_ENUM(N, OPERATOR_APPLY, n)); \
	} \
};

// functor class definition
BOOST_PP_REPEAT(OPERATOR_NUM, OPERATOR_TEMPLATE, _)

#undef OPERATOR_APPLY
#undef OPERATOR_TEMPLATE

} // namespace detail

// TODO: consider using result type of target operator
#define OPERATOR_TEMPLATE(z, n, data) \
	template<typename T, TYPENAME_UN> \
	friend typename detail::OPERATOR_CLASS_N(n)<T, TYPE_UN>::result_type \
	operator OPERATOR_OP(n) (const holderN<TYPE_UN>& u, const T& t); \
	\
	template<typename T, TYPENAME_UN> \
	friend typename detail::OPERATOR_CLASS_N(n)<T, TYPE_UN>::result_type \
	operator OPERATOR_OP_REV(n) (const T& t, const holderN<TYPE_UN>& u);

#define ANYCHECK_TEMPLATE(z, n, data) || boost::get<n>(tuple)
#define ALLCHECK_TEMPLATE(z, n, data) && boost::get<n>(tuple)

//#define CONST_TN_REF_(z, n, data) const T##n &
//#define CONST_TN_REF BOOST_PP_ENUM(N, CONST_TN_REF_, _)

template<TYPENAME_TN>
class holderN
{
public:
	holderN(spos_type t, CONST_TN_REF_TN) :
		type(t), tuple(VAR_TN) {}
	operator bool() const
	{
		if(type == any_type) return any_check();
		else return all_check();
	}
	bool operator!() const
	{
		return ! operator bool();
	}
	// friend operator declaration
	BOOST_PP_REPEAT(OPERATOR_NUM, OPERATOR_TEMPLATE, _)
private:
	spos_type type;
// results of operator application are temporary.
// Using const reference may cause unclear mysterious failures.
//	boost::tuple<CONST_TN_REF> tuple;
	boost::tuple<TYPE_TN> tuple;
	holderN& operator=(const holderN&);
	bool any_check() const
	{
		return false BOOST_PP_REPEAT(N, ANYCHECK_TEMPLATE, _);
	}
	bool all_check() const
	{
		return true BOOST_PP_REPEAT(N, ALLCHECK_TEMPLATE, _);
	}
};

#undef OPERATOR_TEMPLATE
#undef ANYCHECK_TEMPLATE
#undef ALLCHECK_TEMPLATE

#define CALL(z, n, data) boost::get<n>(u.tuple)

// template for friend operator definition
#define OPERATOR_TEMPLATE(z, n, data) \
template<typename T, TYPENAME_UN> \
typename detail::OPERATOR_CLASS_N(n)<T, TYPE_UN>::result_type operator OPERATOR_OP(n) (const holderN<TYPE_UN>& u, const T& t) \
{ \
	detail::OPERATOR_CLASS_N(n)<T, TYPE_UN> op(u.type, t); \
	return op(BOOST_PP_ENUM(N, CALL, _)); \
} \
\
template<typename T, TYPENAME_UN> \
typename detail::OPERATOR_CLASS_N(n)<T, TYPE_UN>::result_type operator OPERATOR_OP_REV(n) (const T& t, const holderN<TYPE_UN>& u) \
{ \
	detail::OPERATOR_CLASS_N(n)<T, TYPE_UN> op(u.type, t); \
	return op(BOOST_PP_ENUM(N, CALL, _)); \
}

// friend operator definition
BOOST_PP_REPEAT(OPERATOR_NUM, OPERATOR_TEMPLATE, _)

#undef CALL
#undef OPERATOR_TEMPLATE

template<TYPENAME_TN>
holderN<TYPE_TN>
any(CONST_TN_REF_TN)
{
	return holderN<TYPE_TN>(any_type, VAR_TN);
}

template<TYPENAME_TN>
holderN<TYPE_TN>
all(CONST_TN_REF_TN)
{
	return holderN<TYPE_TN>(all_type, VAR_TN);
}

#undef N

#undef TYPE_TN
#undef VAR_TN
#undef TYPE_UN
#undef VAR_UN
#undef TYPENAME_TN
#undef TYPENAME_UN
#undef CONST_TN_REF_TN

#endif // !BOOST_PP_IS_ITERATING
