/*****************************************************************************/
//
//  superposition_impl_v.hpp: C++ version of Perl module "Quantum::Superpositions"
//      Implementation for variadic templates version
//
//   $URL: https://yak.myhome.cx/repos/source/mylib/trunk/cpp/superposition_impl_v.hpp $
//  $Date: 2007-10-25 01:08:28 +0900 (Thu, 25 Oct 2007) $
//   $Rev: 478 $
//
//      Copyright (c) 2007 Yak!
//
//  This code is distributed under the zlib/libpng License.
//  See http://www.opensource.org/licenses/zlib-license.php
//
/*****************************************************************************/

#ifndef YAK_UTIL_SUPERPOSITION_V_HPP
#define YAK_UTIL_SUPERPOSITION_V_HPP

#include <boost/mpl/always.hpp>

// Include boost or std tuple also, depneding on whether YAK_USE_BOOST_TUPLE is defined or not.
// Corresponding get() can be used via yak::util::detail::get()
#include "variadic_helper.hpp"

namespace yak {
namespace util {
namespace superposition {

// forward declaration

// should prohibit no type parameter
template<typename ... Tn>
class holder;

namespace detail {

// Conversion to bool

// TODO: safe bool idiom should be used.

// For partial specialization, we use functor class instead of function template.

// for "any" to bool

template<int N, typename ... Tn>
struct any_check
{
	bool operator()(const typename variadic_to_tuple<Tn...>::type & tuple)
	{
		return yak::util::detail::get<N>(tuple) || any_check<N-1, Tn...>()(tuple);
	}
};


template<typename ... Tn>
struct any_check<0, Tn...>
{
	bool operator()(const typename variadic_to_tuple<Tn...>::type & tuple)
	{
		return yak::util::detail::get<0>(tuple);
	}
};

// for "all" to bool

template<int N, typename ... Tn>
struct all_check
{
	bool operator()(const typename variadic_to_tuple<Tn...>::type & tuple)
	{
		return yak::util::detail::get<N>(tuple) && all_check<N-1, Tn...>()(tuple);
	}
};

template<typename ... Tn>
struct all_check<0, Tn...>
{
	bool operator()(const typename variadic_to_tuple<Tn...>::type & tuple)
	{
		return yak::util::detail::get<0>(tuple);
	}
};

// Functor class for operator

template<typename U, typename ... Tn>
struct RelOp
{
	typedef holder<typename boost::mpl::always<bool>::template apply<Tn>::type ...> result_type;
	RelOp(spos_type type, const U& u) : type(type), arg(u) {}
	spos_type type;
	const U& arg;
};

// template for functor class defniition
#define OPERATOR_TEMPLATE(z, n, data) \
template<typename U, typename ... Tn> \
struct OPERATOR_CLASS(n) : public RelOp<U, Tn...> \
{ \
	typedef typename RelOp<U, Tn...>::result_type result_type; \
	OPERATOR_CLASS(n)(spos_type type, const U& u) : RelOp<U, Tn...>(type, u) {} \
	result_type operator()(const Tn& ... tn) \
	{ \
		return result_type(this->type, (tn OPERATOR_OP(n) this->arg)...); \
	} \
};

// functor class definition
BOOST_PP_REPEAT(OPERATOR_NUM, OPERATOR_TEMPLATE, _)

#undef OPERATOR_TEMPLATE

} // namespace detail

// TODO: consider using result type of target operator
#define OPERATOR_TEMPLATE(z, n, data) \
	template<typename T, typename ... Un> \
	friend typename detail::OPERATOR_CLASS(n)<T, Un...>::result_type \
	operator OPERATOR_OP(n) (const holder<Un...>& u, const T& t); \
	\
	template<typename T, typename ... Un> \
	friend typename detail::OPERATOR_CLASS(n)<T, Un...>::result_type \
	operator OPERATOR_OP_REV(n) (const T& t, const holder<Un...>& u);

template<typename ... Tn>
class holder
{
public:
	holder(spos_type t, const Tn& ... tn) : type(t), tuple(tn...) {}
	operator bool() const
	{
		if(type == any_type) return detail::any_check<sizeof...(Tn)-1, Tn...>()(tuple);
		else return detail::all_check<sizeof...(Tn)-1, Tn...>()(tuple);
	}
	bool operator!() const
	{
		return ! operator bool();
	}
	// friend operator declaration
	BOOST_PP_REPEAT(OPERATOR_NUM, OPERATOR_TEMPLATE, _)
private:
	spos_type type;
// results of operator application are temporary.
// Using const reference may cause unclear mysterious failures.
//	typename variadic_to_tuple<const Tn&...>::type tuple;
	typename variadic_to_tuple<Tn...>::type tuple;
	holder& operator=(const holder&);
};

#undef OPERATOR_TEMPLATE

// template for friend operator definition
#define OPERATOR_TEMPLATE(z, n, data) \
template<typename T, typename ... Un> \
typename detail::OPERATOR_CLASS(n)<T, Un...>::result_type operator OPERATOR_OP(n) (const holder<Un...>& u, const T& t) \
{ \
	detail::OPERATOR_CLASS(n)<T, Un...> op(u.type, t); \
	return tuple_to_variadic<typename detail::OPERATOR_CLASS(n)<T, Un...>::result_type>(u.tuple, op); \
} \
\
template<typename T, typename ... Un> \
typename detail::OPERATOR_CLASS(n)<T, Un...>::result_type operator OPERATOR_OP_REV(n) (const T& t, const holder<Un...>& u) \
{ \
	detail::OPERATOR_CLASS(n)<T, Un...> op(u.type, t); \
	return tuple_to_variadic<typename detail::OPERATOR_CLASS(n)<T, Un...>::result_type>(u.tuple, op); \
}

// friend operator definition
BOOST_PP_REPEAT(OPERATOR_NUM, OPERATOR_TEMPLATE, _)

#undef OPERATOR_TEMPLATE

template<typename ... Tn>
holder<Tn...>
any(const Tn& ... tn)
{
	return holder<Tn...>(any_type, tn...);
}

template<typename ... Tn>
holder<Tn...>
all(const Tn& ... tn)
{
	return holder<Tn...>(all_type, tn...);
}

} // namespace superposition
} // namespace util
} // namesapce yak

#endif // YAK_UTIL_SUPERPOSITION_V_HPP
